from collections.abc import Mapping
from contextlib import contextmanager
from ctypes import GetLastError, WinError, byref, create_unicode_buffer, sizeof
from typing import Any, Iterable, Iterator, TypeVar

from winerror import ERROR_NO_MORE_ITEMS

from winpnp._setupapi import (
    DIGCF,
    HDEVINFO,
    INVALID_HANDLE_VALUE,
    SP_DEVINFO_DATA,
    SetupDiCreateDeviceInfoList,
    SetupDiDestroyDeviceInfoList,
    SetupDiEnumDeviceInfo,
    SetupDiGetClassDevsW,
    SetupDiGetDevicePropertyKeys,
    SetupDiGetDevicePropertyW,
    SetupDiOpenDeviceInfoW,
)
from winpnp.properties import keys
from winpnp.properties.pnp_property import PnpProperty, PnpPropertyKey

from ._pnp_property_mapping import PnpPropertyMapping

T = TypeVar("T")


class DeviceInfo(Mapping[PnpPropertyKey[Any], PnpProperty[Any]]):
    """
    Represents information about a single PnP device node.

    This class should not be instantiated directly. Instead, use one of the factory methods: `of_instance_id`, `of_all_devices`, `of_present_devices`:

    Can be used as a `Mapping`, with `PnpPropertyKey`s as keys and `PnpProperty` as values.

    For example:
    >>> from winpnp.properties.keys.device import INSTANCE_ID
    ...
    >>> with DeviceInfo.of_instance_id("HTREE\\ROOT\\0") as device:
    ...     instance_id = device[INSTANCE_ID]
    ...
    >>> instance_id
    PnpProperty(value='HTREE\\ROOT\\0', kind=PnpPropertyType(type_id=18, name='STRING'))
    """

    __init_token = object()

    def __init__(self, handle: HDEVINFO, data: SP_DEVINFO_DATA, *, _token=None) -> None:
        """
        This function should not be called directly, instead use one of the factory methods: `of_instance_id`, `of_all_devices`, `of_present_devices`.
        """
        if _token is not self.__init_token:
            raise TypeError(
                f"{type(self).__name__} should not be instantiated directly. You should use one of the factory methods: `of_instance_id`, `of_all_devices`, `of_present_devices`"
            )

        super().__init__()

        self._handle = handle
        self._data = data

        self._properties: PnpPropertyMapping[Any] = PnpPropertyMapping(
            description="<uninitialized DeviceInfo>",
            query_value_func=lambda PropertyKey, PropertyType, PropertyBuffer, PropertyBufferSize, RequiredSize: SetupDiGetDevicePropertyW(
                self._handle,
                byref(self._data),
                PropertyKey,
                PropertyType,
                PropertyBuffer,
                PropertyBufferSize,
                RequiredSize,
                0,
            ),
            query_keys_func=lambda PropertyKeyArray, PropertyKeyCount, RequiredPropertyKeyCount: SetupDiGetDevicePropertyKeys(
                self._handle,
                byref(self._data),
                PropertyKeyArray,
                PropertyKeyCount,
                RequiredPropertyKeyCount,
                0,
            ),
        )

        self._instance_id: str = self._properties[keys.device.INSTANCE_ID].value
        # Update description now that instance_id is known
        self._properties.description = repr(self)

    def __repr__(self) -> str:
        return f"{type(self).__name__}(instance_id={repr(self._instance_id)})"

    def __getitem__(self, key: PnpPropertyKey[T]) -> PnpProperty[T]:
        return self._properties[key]

    def __len__(self) -> int:
        return len(self._properties)

    def __iter__(self) -> Iterator[PnpPropertyKey[Any]]:
        return iter(self._properties)

    @classmethod
    def __enumerate_device_info_set(cls, handle: HDEVINFO) -> Iterator["DeviceInfo"]:
        index = 0
        data = SP_DEVINFO_DATA()
        data.cbSize = sizeof(data)

        while SetupDiEnumDeviceInfo(handle, index, byref(data)):
            yield cls(
                handle, SP_DEVINFO_DATA.from_buffer_copy(data), _token=cls.__init_token
            )
            index += 1

        error = GetLastError()
        if error != ERROR_NO_MORE_ITEMS:
            raise WinError(error)

    @classmethod
    @contextmanager
    def __create_using_SetupDiGetClassDevsW(
        cls, flags: DIGCF
    ) -> Iterator[Iterable["DeviceInfo"]]:
        handle: HDEVINFO = SetupDiGetClassDevsW(None, None, None, flags)
        if handle == INVALID_HANDLE_VALUE:
            raise WinError()

        try:
            yield cls.__enumerate_device_info_set(handle)
        finally:
            SetupDiDestroyDeviceInfoList(handle)

    @classmethod
    @contextmanager
    def of_instance_id(cls, instance_id: str) -> Iterator["DeviceInfo"]:
        """
        Opens information of the device with the specified instance id.
        """
        handle: HDEVINFO = SetupDiCreateDeviceInfoList(None, None)
        if handle == INVALID_HANDLE_VALUE:
            raise WinError()

        try:
            data = SP_DEVINFO_DATA()
            data.cbSize = sizeof(data)
            if not SetupDiOpenDeviceInfoW(
                handle, create_unicode_buffer(instance_id), None, 0, byref(data)
            ):
                raise WinError()

            yield cls(handle, data, _token=cls.__init_token)
        finally:
            SetupDiDestroyDeviceInfoList(handle)

    @classmethod
    @contextmanager
    def of_all_devices(cls) -> Iterator[Iterable["DeviceInfo"]]:
        """
        Opens information of all devices known to Windows, including ones that are no longer present.
        """
        with cls.__create_using_SetupDiGetClassDevsW(DIGCF.ALLCLASSES) as devices:
            yield devices

    @classmethod
    @contextmanager
    def of_present_devices(cls) -> Iterator[Iterable["DeviceInfo"]]:
        """
        Opens information of all devices that are currently present.
        """
        with cls.__create_using_SetupDiGetClassDevsW(
            DIGCF.ALLCLASSES | DIGCF.PRESENT
        ) as devices:
            yield devices



from uuid import UUID

from winpnp.properties import kinds
from winpnp.properties.pnp_property import PnpPropertyKey

FRIENDLY_NAME = PnpPropertyKey.register_new(
    UUID("{026e516e-b814-414b-83cd856d6fef4822}"),
    2,
    "DeviceInterface_FriendlyName",
    (kinds.STRING,),
)

ENABLED = PnpPropertyKey.register_new(
    UUID("{026e516e-b814-414b-83cd856d6fef4822}"),
    3,
    "DeviceInterface_Enabled",
    (kinds.BOOLEAN,),
)

CLASS_GUID = PnpPropertyKey.register_new(
    UUID("{026e516e-b814-414b-83cd856d6fef4822}"),
    4,
    "DeviceInterface_ClassGuid",
    (kinds.GUID,),
)

REFERENCE_STRING = PnpPropertyKey.register_new(
    UUID("{026e516e-b814-414b-83cd856d6fef4822}"),
    5,
    "DeviceInterface_ReferenceString",
    (kinds.STRING,),
)

RESTRICTED = PnpPropertyKey.register_new(
    UUID("{026e516e-b814-414b-83cd856d6fef4822}"),
    6,
    "DeviceInterface_Restricted",
    (kinds.BOOLEAN,),
)

UNRESTRICTED_APP_CAPABILITIES = PnpPropertyKey.register_new(
    UUID("{026e516e-b814-414b-83cd856d6fef4822}"),
    8,
    "DeviceInterface_UnrestrictedAppCapabilities",
    (kinds.STRING_LIST,),
)

SCHEMATIC_NAME = PnpPropertyKey.register_new(
    UUID("{026e516e-b814-414b-83cd856d6fef4822}"),
    9,
    "DeviceInterface_SchematicName",
    (kinds.STRING,),
)



from uuid import UUID

from winpnp.properties import kinds
from winpnp.properties.pnp_property import PnpPropertyKey

MODEL = PnpPropertyKey.register_new(
    UUID("{cf73bb51-3abf-44a2-85e09a3dc7a12132}"), 2, "DrvPkg_Model", (kinds.STRING,)
)

VENDOR_WEB_SITE = PnpPropertyKey.register_new(
    UUID("{cf73bb51-3abf-44a2-85e09a3dc7a12132}"),
    3,
    "DrvPkg_VendorWebSite",
    (kinds.STRING,),
)

DETAILED_DESCRIPTION = PnpPropertyKey.register_new(
    UUID("{cf73bb51-3abf-44a2-85e09a3dc7a12132}"),
    4,
    "DrvPkg_DetailedDescription",
    (kinds.STRING,),
)

DOCUMENTATION_LINK = PnpPropertyKey.register_new(
    UUID("{cf73bb51-3abf-44a2-85e09a3dc7a12132}"),
    5,
    "DrvPkg_DocumentationLink",
    (kinds.STRING,),
)

ICON = PnpPropertyKey.register_new(
    UUID("{cf73bb51-3abf-44a2-85e09a3dc7a12132}"),
    6,
    "DrvPkg_Icon",
    (kinds.STRING_LIST,),
)

BRANDING_ICON = PnpPropertyKey.register_new(
    UUID("{cf73bb51-3abf-44a2-85e09a3dc7a12132}"),
    7,
    "DrvPkg_BrandingIcon",
    (kinds.STRING_LIST,),
)



from collections.abc import Mapping
from ctypes import WinError, byref, c_uint32, create_unicode_buffer
from typing import Any, Iterator, TypeVar
from uuid import UUID

from winpnp._setupapi import (
    DICLASSPROP,
    GUID,
    SetupDiBuildClassInfoList,
    SetupDiClassGuidsFromNameW,
    SetupDiGetClassPropertyKeys,
    SetupDiGetClassPropertyW,
)
from winpnp.properties import keys
from winpnp.properties.pnp_property import PnpProperty, PnpPropertyKey

from ._pnp_property_mapping import PnpPropertyMapping

T = TypeVar("T")


class SetupClassInfo(Mapping[PnpPropertyKey[Any], PnpProperty[Any]]):
    """
    Represents information about a single PnP setup class.

    Can be used as a `Mapping`, with `PnpPropertyKey`s as keys and `PnpProperty` as values.

    For example:
    >>> from uuid import UUID
    >>> from winpnp.properties.keys.device_class import CLASS_NAME
    ...
    >>> SetupClassInfo(UUID("36fc9e60-c465-11cf-8056-444553540000"))[CLASS_NAME]
    PnpProperty(value='USB', kind=PnpPropertyType(type_id=18, name='STRING'))
    """

    def __init__(self, guid: UUID) -> None:
        super().__init__()

        self._guid = guid

        self._properties: PnpPropertyMapping[Any] = PnpPropertyMapping(
            description="<uninitialized SetupClassInfo>",
            query_value_func=lambda PropertyKey, PropertyType, PropertyBuffer, PropertyBufferSize, RequiredSize: SetupDiGetClassPropertyW(
                byref(GUID.from_uuid(self._guid)),
                PropertyKey,
                PropertyType,
                PropertyBuffer,
                PropertyBufferSize,
                RequiredSize,
                DICLASSPROP.INSTALLER,
            ),
            query_keys_func=lambda PropertyKeyArray, PropertyKeyCount, RequiredPropertyKeyCount: SetupDiGetClassPropertyKeys(
                byref(GUID.from_uuid(self._guid)),
                PropertyKeyArray,
                PropertyKeyCount,
                RequiredPropertyKeyCount,
                DICLASSPROP.INSTALLER,
            ),
        )

        self._name: str = self._properties[keys.device_class.CLASS_NAME].value
        # Update description now that instance_id is known
        self._properties.description = repr(self)

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}(guid={repr(self._guid)}, name={repr(self._name)})"
        )

    def __getitem__(self, key: PnpPropertyKey[T]) -> PnpProperty[T]:
        return self._properties[key]

    def __len__(self) -> int:
        return len(self._properties)

    def __iter__(self) -> Iterator[PnpPropertyKey]:
        return iter(self._properties)

    @property
    def guid(self) -> UUID:
        return self._guid

    @classmethod
    def of_class_name(cls, name: str) -> Iterator["SetupClassInfo"]:
        name_buffer = create_unicode_buffer(name)
        required_size = c_uint32()
        if SetupDiClassGuidsFromNameW(name_buffer, None, 0, byref(required_size)):
            return

        guids = (GUID * required_size.value)()
        if not SetupDiClassGuidsFromNameW(
            name_buffer, guids, required_size, byref(required_size)
        ):
            raise WinError()

        for guid in guids:
            yield cls(guid.to_uuid())

    @classmethod
    def of_all_classes(cls) -> Iterator["SetupClassInfo"]:
        required_size = c_uint32()
        if SetupDiBuildClassInfoList(0, None, 0, byref(required_size)):
            return

        guids = (GUID * required_size.value)()
        if not SetupDiBuildClassInfoList(0, guids, required_size, byref(required_size)):
            raise WinError()

        for guid in guids:
            yield cls(guid.to_uuid())



from dataclasses import dataclass, field
from typing import Any, Callable, ClassVar, Generic, Iterable, Optional, TypeVar
from uuid import UUID

from .decoding import decode_raw

T = TypeVar("T")
U = TypeVar("U")


@dataclass(init=False)
class PnpPropertyType(Generic[T]):
    """
    Represents a type of PnP properties. This is an abstraction over the Windows DEVPROP_TYPE constants.
    """

    _DERIVED_DATA: ClassVar[dict[int, tuple[str, Callable[[bytes], Any]]]] = {}

    type_id: int
    name: Optional[str] = field(compare=False)
    _decoder: Callable[[bytes], T] = field(repr=False, compare=False)

    def __init__(
        self,
        type_id: int,
        name: Optional[str] = None,
        decoder: Optional[Callable[[bytes], T]] = None,
    ) -> None:
        derived_name, derived_decoder = self._DERIVED_DATA.get(
            type_id, (None, decode_raw)
        )

        self.type_id = type_id
        self.name = name if name is not None else derived_name
        self._decoder = decoder if decoder is not None else derived_decoder

    @staticmethod
    def register(kind: "PnpPropertyType[U]") -> None:
        """
        Registers the name and decoder of the specified `PnpPropertyType` as the default name and decoder for other instances with the same `type_id`.
        """
        if kind.name is None:
            raise ValueError(f"Cannot register {repr(kind)} because it is unnamed.")

        if kind.type_id in PnpPropertyType._DERIVED_DATA:
            raise ValueError(
                f"Cannot register {repr(kind)} because its type_id is already registered with name {PnpPropertyType._DERIVED_DATA[kind.type_id][0]}"
            )

        PnpPropertyType._DERIVED_DATA[kind.type_id] = (kind.name, kind._decoder)

    @staticmethod
    def register_new(
        type_id: int,
        name: str,
        decoder: Callable[[bytes], U],
    ) -> "PnpPropertyType[U]":
        """
        Creates a new `PnpPropertyType`, registers it, and returns it.
        """
        kind = PnpPropertyType(type_id, name, decoder)
        PnpPropertyType.register(kind)
        return kind

    def decode(self, data: bytes) -> T:
        """
        Decodes raw bytes to the actual python type that this `PnpPropertyType` represents.
        """
        return self._decoder(data)


@dataclass(init=False)
class PnpPropertyKey(Generic[T]):
    """
    A key that specifies a PnP property. This is an abstraction over the Windows DEVPROPKEY struct. Can be used as key for `__getitem__` of various winpnp classes.
    """

    _NAMES: ClassVar[dict[tuple[UUID, int], str]] = {}

    category: UUID
    property_id: int
    name: Optional[str] = field(compare=False)
    allowed_types: Optional[dict[int, PnpPropertyType[T]]] = field(compare=False)

    def __init__(
        self,
        category: UUID,
        property_id: int,
        name: Optional[str] = None,
        allowed_types: Optional[Iterable[PnpPropertyType[T]]] = None,
    ) -> None:
        self.category = category
        self.property_id = property_id
        self.name = (
            name if name is not None else self._NAMES.get((category, property_id))
        )
        self.allowed_types = (
            {kind.type_id: kind for kind in allowed_types}
            if allowed_types is not None
            else None
        )

    @staticmethod
    def register(key: "PnpPropertyKey[U]") -> None:
        """
        Registers the name and of the specified `PnpPropertyKey` as the default name for other instances with the same `(category, property_id) pair`.
        """
        if key.name is None:
            raise ValueError(f"Cannot register {repr(key)} because it is unnamed.")

        _id = (key.category, key.property_id)
        if _id in PnpPropertyKey._NAMES:
            raise ValueError(
                f"Cannot register {repr(key)} because its (category, property_id) pair is already registered with name {repr(PnpPropertyKey._NAMES[_id])}"
            )

        PnpPropertyKey._NAMES[_id] = key.name

    @staticmethod
    def register_new(
        category: UUID,
        property_id: int,
        name: str,
        allowed_types: Optional[Iterable[PnpPropertyType[U]]] = None,
    ) -> "PnpPropertyKey[U]":
        """
        Creates a new `PnpPropertyKey`, registers it, and returns it.
        """
        key = PnpPropertyKey(category, property_id, name, allowed_types)
        PnpPropertyKey.register(key)
        return key


@dataclass()
class PnpProperty(Generic[T]):
    """
    Holds the value and type of a PnP property.
    """

    value: T
    kind: PnpPropertyType[T]



from uuid import UUID

from winpnp.properties import kinds
from winpnp.properties.pnp_property import PnpPropertyKey

DEFAULT_INTERFACE = PnpPropertyKey.register_new(
    UUID("{14c83a99-0b3f-44b7-be4ca178d3990564}"),
    2,
    "DeviceInterfaceClass_DefaultInterface",
    (kinds.STRING,),
)

NAME = PnpPropertyKey.register_new(
    UUID("{14c83a99-0b3f-44b7-be4ca178d3990564}"),
    3,
    "DeviceInterfaceClass_Name",
    (kinds.STRING,),
)



import ctypes
from array import array
from ctypes import FormatError, Structure, byref, c_int32, c_uint16, sizeof, windll
from dataclasses import dataclass, field
from datetime import datetime, timezone, tzinfo
from itertools import takewhile
from typing import TYPE_CHECKING, Any, Literal, Optional
from uuid import UUID

from more_itertools import sliced

from winpnp import _setupapi

if TYPE_CHECKING:
    # Import only if type checking to avoid circular import
    from .pnp_property import PnpPropertyKey, PnpPropertyType

_GUID_BYTE_SIZE = 16
_FILETIME_BYTE_SIZE = 8

_ArrayIntTypeCode = Literal["b", "B", "h", "H", "i", "I", "l", "L", "q", "Q"]
_ArrayFloatTypeCode = Literal["f", "d"]


class _SYSTEMTIME(Structure):
    _fields_ = (
        ("wYear", c_uint16),
        ("wMonth", c_uint16),
        ("wDayOfWeek", c_uint16),
        ("wDay", c_uint16),
        ("wHour", c_uint16),
        ("wMinute", c_uint16),
        ("wSecond", c_uint16),
        ("wMilliseconds", c_uint16),
    )

    def to_datetime(self, time_zone: Optional[tzinfo] = None) -> datetime:
        return datetime(
            self.wYear,
            self.wMonth,
            self.wDay,
            self.wHour,
            self.wMinute,
            self.wSecond,
            self.wMilliseconds * 1000,
            time_zone,
        )


@dataclass(frozen=True)
class Win32Error:
    code: int
    description: str = field(init=False, compare=False)

    def __post_init__(self) -> None:
        # using object.__setattr__ because self is frozen
        object.__setattr__(self, "description", FormatError(c_int32(self.code).value))


@dataclass(frozen=True)
class NtStatus:
    code: int
    description: str = field(init=False, compare=False)

    def __post_init__(self) -> None:
        # using object.__setattr__ because self is frozen
        object.__setattr__(
            self,
            "description",
            FormatError(c_int32(windll.ntdll.RtlNtStatusToDosError(self.code)).value),
        )


def decode_raw(data: bytes) -> bytes:
    return data


def decode_integers(data: bytes, typecode: _ArrayIntTypeCode) -> list[int]:
    return list(array(typecode, data))


def decode_floats(data: bytes, typecode: _ArrayFloatTypeCode) -> list[float]:
    return list(array(typecode, data))


def decode_guids(data: bytes) -> list[UUID]:
    return [UUID(bytes_le=x) for x in sliced(data, _GUID_BYTE_SIZE, strict=True)]


def decode_filetimes(data: bytes) -> list[datetime]:
    output = []
    systemtime = _SYSTEMTIME()

    for filetime in sliced(data, _FILETIME_BYTE_SIZE, strict=True):
        if not windll.kernel32.FileTimeToSystemTime(filetime, byref(systemtime)):
            raise ctypes.WinError()
        output.append(systemtime.to_datetime(timezone.utc))

    return output


def decode_booleans(data: bytes) -> list[bool]:
    return [x != 0 for x in data]


def decode_string(data: bytes) -> str:
    string = data.decode("utf-16-le")
    return "".join(takewhile(lambda char: char != "\0", string))


def decode_strings(data: bytes) -> list[str]:
    strings = data.decode("utf-16-le").split("\0")
    return list(takewhile(lambda x: len(x) > 0, strings))


def decode_property_keys(data: bytes) -> list["PnpPropertyKey[Any]"]:
    # Import in function to avoid circular import
    from .pnp_property import PnpPropertyKey

    count = len(data) // sizeof(_setupapi.DEVPROPKEY)
    keys = (_setupapi.DEVPROPKEY * count).from_buffer_copy(data)
    return [PnpPropertyKey(key.fmtid.to_uuid(), key.pid) for key in keys]


def decode_property_types(data: bytes) -> list["PnpPropertyType[Any]"]:
    # Import in function to avoid circular import
    from .pnp_property import PnpPropertyType

    return [PnpPropertyType(type_id) for type_id in decode_integers(data, "L")]


def decode_win32_errors(data: bytes) -> list[Win32Error]:
    return [Win32Error(code) for code in array("L", data)]


def decode_nt_statuses(data: bytes) -> list[NtStatus]:
    return [NtStatus(code) for code in array("L", data)]



from ctypes import (
    POINTER,
    Structure,
    c_bool,
    c_ubyte,
    c_uint16,
    c_uint32,
    c_void_p,
    c_wchar_p,
    windll,
)
from enum import IntFlag
from uuid import UUID

INVALID_HANDLE_VALUE = -1

HDEVINFO = c_void_p


class DIGCF(IntFlag):
    DEFAULT = 0x00000001
    PRESENT = 0x00000002
    ALLCLASSES = 0x00000004
    PROFILE = 0x00000008
    DEVICEINTERFACE = 0x00000010


class DICLASSPROP(IntFlag):
    INSTALLER = 0x00000001
    INTERFACE = 0x00000002


class GUID(Structure):
    _fields_ = (
        ("Data1", c_uint32),
        ("Data2", c_uint16),
        ("Data3", c_uint16),
        ("Data4", c_ubyte * 8),
    )

    def __str__(self) -> str:
        return str(self.to_uuid())

    def __repr__(self) -> str:
        return f"{type(self).__name__}({{{str(self.to_uuid())}}})"

    @classmethod
    def from_uuid(cls, guid: UUID) -> "GUID":
        return cls.from_buffer_copy(guid.bytes_le)

    def to_uuid(self) -> UUID:
        return UUID(bytes_le=bytes(self))


class SP_DEVINFO_DATA(Structure):
    _fields_ = (
        ("cbSize", c_uint32),
        ("ClassGuid", GUID),
        ("DevInst", c_uint32),
        ("Reserved", c_void_p),
    )


class DEVPROPKEY(Structure):
    _fields_ = (
        ("fmtid", GUID),
        ("pid", c_uint32),
    )


class DEVPROP_TYPEMOD(IntFlag):
    NONE = 0x00000000
    ARRAY = 0x00001000  # array of fixed-sized data elements
    LIST = 0x00002000  # list of variable-sized data elements


class DEVPROP_TYPE(IntFlag):
    EMPTY = 0x00000000  # nothing, no property data
    NULL = 0x00000001  # null property data
    SBYTE = 0x00000002  # 8-bit signed int (SBYTE)
    BYTE = 0x00000003  # 8-bit unsigned int (BYTE)
    INT16 = 0x00000004  # 16-bit signed int (SHORT)
    UINT16 = 0x00000005  # 16-bit unsigned int (USHORT)
    INT32 = 0x00000006  # 32-bit signed int (LONG)
    UINT32 = 0x00000007  # 32-bit unsigned int (ULONG)
    INT64 = 0x00000008  # 64-bit signed int (LONG64)
    UINT64 = 0x00000009  # 64-bit unsigned int (ULONG64)
    FLOAT = 0x0000000A  # 32-bit floating-point (FLOAT)
    DOUBLE = 0x0000000B  # 64-bit floating-point (DOUBLE)
    DECIMAL = 0x0000000C  # 128-bit data (DECIMAL)
    GUID = 0x0000000D  # 128-bit unique identifier (GUID)
    CURRENCY = 0x0000000E  # 64 bit signed int currency value (CURRENCY)
    DATE = 0x0000000F  # date (DATE)
    FILETIME = 0x00000010  # file time (FILETIME)
    BOOLEAN = 0x00000011  # 8-bit boolean (DEVPROP_BOOLEAN)
    STRING = 0x00000012  # null-terminated string
    STRING_LIST = STRING | DEVPROP_TYPEMOD.LIST  # multi-sz string list
    SECURITY_DESCRIPTOR = 0x00000013  # self-relative binary SECURITY_DESCRIPTOR
    SECURITY_DESCRIPTOR_STRING = 0x00000014  # security descriptor string (SDDL format)
    DEVPROPKEY = 0x00000015  # device property key (DEVPROPKEY)
    DEVPROPTYPE = 0x00000016  # device property type (DEVPROPTYPE)
    BINARY = BYTE | DEVPROP_TYPEMOD.ARRAY  # custom binary data
    ERROR = 0x00000017  # 32-bit Win32 system error code
    NTSTATUS = 0x00000018  # 32-bit NTSTATUS code
    STRING_INDIRECT = 0x00000019  # string resource (@[path\]<dllname>,-<strId>)


SetupDiCreateDeviceInfoList = windll.setupapi.SetupDiCreateDeviceInfoList
SetupDiCreateDeviceInfoList.restype = HDEVINFO
SetupDiCreateDeviceInfoList.argtypes = (
    POINTER(GUID),  # [in, optional] ClassGuid
    c_void_p,  # [in, optional] hwndParent
)

SetupDiDestroyDeviceInfoList = windll.setupapi.SetupDiDestroyDeviceInfoList
SetupDiDestroyDeviceInfoList.restype = c_bool
SetupDiDestroyDeviceInfoList.argtypes = (HDEVINFO,)  # [in] DeviceInfoSet

SetupDiGetClassDevsW = windll.setupapi.SetupDiGetClassDevsW
SetupDiGetClassDevsW.restype = HDEVINFO
SetupDiGetClassDevsW.argtypes = (
    POINTER(GUID),  # [in, optional] ClassGuid
    c_wchar_p,  # [in, optional] Enumerator
    c_void_p,  # [in, optional] hwndParent
    c_uint32,  # [in] Flags
)

SetupDiEnumDeviceInfo = windll.setupapi.SetupDiEnumDeviceInfo
SetupDiEnumDeviceInfo.restype = c_bool
SetupDiEnumDeviceInfo.argtypes = (
    HDEVINFO,  # [in] DeviceInfoSet
    c_uint32,  # [in] MemberIndex
    POINTER(SP_DEVINFO_DATA),  # [out] DeviceInfoData
)

SetupDiOpenDeviceInfoW = windll.setupapi.SetupDiOpenDeviceInfoW
SetupDiOpenDeviceInfoW.restype = c_bool
SetupDiOpenDeviceInfoW.argtypes = (
    HDEVINFO,  # [in] DeviceInfoSet
    c_wchar_p,  # [in] DeviceInstanceId
    c_void_p,  # [in, optional] hwndParent
    c_uint32,  # [in] OpenFlags
    POINTER(SP_DEVINFO_DATA),  # [out, optional] DeviceInfoData
)

SetupDiGetDevicePropertyW = windll.setupapi.SetupDiGetDevicePropertyW
SetupDiGetDevicePropertyW.restype = c_bool
SetupDiGetDevicePropertyW.argtypes = (
    HDEVINFO,  # [in] DeviceInfoSet
    POINTER(SP_DEVINFO_DATA),  # [in] DeviceInfoData
    POINTER(DEVPROPKEY),  # [in] PropertyKey
    POINTER(c_uint32),  # [out] PropertyType
    POINTER(c_ubyte),  # [out, optional] PropertyBuffer
    c_uint32,  # [in] PropertyBufferSize
    POINTER(c_uint32),  # [out, optional] RequiredSize
    c_uint32,  # [in] Flags
)

SetupDiGetDevicePropertyKeys = windll.setupapi.SetupDiGetDevicePropertyKeys
SetupDiGetDevicePropertyKeys.restype = c_bool
SetupDiGetDevicePropertyKeys.argtypes = (
    HDEVINFO,  # [in] DeviceInfoSet
    POINTER(SP_DEVINFO_DATA),  # [in] DeviceInfoData
    POINTER(DEVPROPKEY),  # [out, optional] PropertyKeyArray
    c_uint32,  # [in] PropertyKeyCount
    POINTER(c_uint32),  # [out, optional] RequiredPropertyKeyCount
    c_uint32,  # [in] Flags
)

SetupDiBuildClassInfoList = windll.setupapi.SetupDiBuildClassInfoList
SetupDiBuildClassInfoList.restype = c_bool
SetupDiBuildClassInfoList.argtypes = (
    c_uint32,  # [in] Flags
    POINTER(GUID),  # [out, optional] ClassGuidList
    c_uint32,  # [in] ClassGuidListSize
    POINTER(c_uint32),  # [out] RequiredSize
)

SetupDiClassGuidsFromNameW = windll.setupapi.SetupDiClassGuidsFromNameW
SetupDiClassGuidsFromNameW.restype = c_bool
SetupDiClassGuidsFromNameW.argtypes = (
    c_wchar_p,  # [in] ClassName,
    POINTER(GUID),  # [out] ClassGuidList,
    c_uint32,  # [in]  ClassGuidListSize,
    POINTER(c_uint32),  # [out] RequiredSize
)

SetupDiGetClassPropertyW = windll.setupapi.SetupDiGetClassPropertyW
SetupDiGetClassPropertyW.restype = c_bool
SetupDiGetClassPropertyW.argtypes = (
    POINTER(GUID),  # [in] ClassGuid,
    POINTER(DEVPROPKEY),  # [in] PropertyKey,
    POINTER(c_uint32),  # [out] PropertyType,
    POINTER(c_ubyte),  # [out] PropertyBuffer,
    c_uint32,  # [in] PropertyBufferSize,
    POINTER(c_uint32),  # [out, optional] RequiredSize,
    c_uint32,  # [in] Flags
)

SetupDiGetClassPropertyKeys = windll.setupapi.SetupDiGetClassPropertyKeys
SetupDiGetClassPropertyKeys.restype = c_bool
SetupDiGetClassPropertyKeys.argtypes = (
    POINTER(GUID),  # [in] ClassGuid,
    POINTER(DEVPROPKEY),  # [out, optional] PropertyKeyArray,
    c_uint32,  # [in] PropertyKeyCount,
    POINTER(c_uint32),  # [out, optional] RequiredPropertyKeyCount,
    c_uint32,  # [in] Flags
)



from more_itertools import one

from winpnp._setupapi import DEVPROP_TYPE, DEVPROP_TYPEMOD

from .decoding import (
    decode_booleans,
    decode_filetimes,
    decode_floats,
    decode_guids,
    decode_integers,
    decode_nt_statuses,
    decode_property_keys,
    decode_property_types,
    decode_raw,
    decode_string,
    decode_strings,
    decode_win32_errors,
)
from .pnp_property import PnpPropertyType

NULL = PnpPropertyType.register_new(int(DEVPROP_TYPE.NULL), "NULL", lambda _: None)
"""The property exists, but it has no value."""

SBYTE = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.SBYTE), "SBYTE", lambda x: one(decode_integers(x, "b"))
)
"""8-bit signed integer"""

SBYTE_ARRAY = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.SBYTE | DEVPROP_TYPEMOD.ARRAY),
    "SBYTE_ARRAY",
    lambda x: decode_integers(x, "b"),
)
"""Array of 8-bit signed integers"""

BYTE = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.BYTE), "BYTE", lambda x: one(decode_integers(x, "B"))
)
"""8-bit unsigned integer"""

BYTE_ARRAY = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.BYTE | DEVPROP_TYPEMOD.ARRAY),
    "BYTE_ARRAY",
    lambda x: decode_integers(x, "B"),
)
"""Array of 8-bit unsigned integers (used for custom binary data)"""

INT16 = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.INT16), "INT16", lambda x: one(decode_integers(x, "h"))
)
"""16-bit signed integer"""

INT16_ARRAY = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.INT16 | DEVPROP_TYPEMOD.ARRAY),
    "INT16_ARRAY",
    lambda x: decode_integers(x, "h"),
)
"""Array of 16-bit signed integers"""

UINT16 = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.UINT16), "UINT16", lambda x: one(decode_integers(x, "H"))
)
"""16-bit unsigned integer"""

UINT16_ARRAY = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.UINT16 | DEVPROP_TYPEMOD.ARRAY),
    "UINT16_ARRAY",
    lambda x: decode_integers(x, "H"),
)
"""Array of 16-bit unsigned integers"""

INT32 = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.INT32), "INT32", lambda x: one(decode_integers(x, "l"))
)
"""32-bit signed integer"""

INT32_ARRAY = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.INT32 | DEVPROP_TYPEMOD.ARRAY),
    "INT32_ARRAY",
    lambda x: decode_integers(x, "l"),
)
"""Array of 32-bit signed integers"""

UINT32 = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.UINT32), "UINT32", lambda x: one(decode_integers(x, "L"))
)
"""32-bit unsigned integer"""

UINT32_ARRAY = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.UINT32 | DEVPROP_TYPEMOD.ARRAY),
    "UINT32_ARRAY",
    lambda x: decode_integers(x, "L"),
)
"""Array of 32-bit unsigned integers"""

INT64 = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.INT64), "INT64", lambda x: one(decode_integers(x, "q"))
)
"""64-bit signed integer"""

INT64_ARRAY = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.INT64 | DEVPROP_TYPEMOD.ARRAY),
    "INT64_ARRAY",
    lambda x: decode_integers(x, "q"),
)
"""Array of 64-bit signed integers"""

UINT64 = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.UINT64), "UINT64", lambda x: one(decode_integers(x, "Q"))
)
"""64-bit unsigned integer"""

UINT64_ARRAY = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.UINT64 | DEVPROP_TYPEMOD.ARRAY),
    "UINT64_ARRAY",
    lambda x: decode_integers(x, "Q"),
)
"""Array of 64-bit unsigned integers"""

FLOAT = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.FLOAT), "FLOAT", lambda x: one(decode_floats(x, "f"))
)
"""32-bit floating-point number"""

FLOAT_ARRAY = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.FLOAT | DEVPROP_TYPEMOD.ARRAY),
    "FLOAT_ARRAY",
    lambda x: decode_floats(x, "f"),
)
"""Array of 32-bit floating-point numbers"""

DOUBLE = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.DOUBLE), "DOUBLE", lambda x: one(decode_floats(x, "d"))
)
"""64-bit floating-point number"""

DOUBLE_ARRAY = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.DOUBLE | DEVPROP_TYPEMOD.ARRAY),
    "DOUBLE_ARRAY",
    lambda x: decode_floats(x, "d"),
)
"""Array of 64-bit floating-point numbers"""

DECIMAL = PnpPropertyType.register_new(int(DEVPROP_TYPE.DECIMAL), "DECIMAL", decode_raw)
"""Windows DECIMAL structure. For now, decoded as raw bytes."""

DECIMAL_ARRAY = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.DECIMAL | DEVPROP_TYPEMOD.ARRAY),
    "DECIMAL_ARRAY",
    decode_raw,
)
"""Array of Windows DECIMAL structures. For now, decoded as raw bytes."""

GUID = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.GUID), "GUID", lambda x: one(decode_guids(x))
)
"""128-bit unique identifier"""

GUID_ARRAY = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.GUID | DEVPROP_TYPEMOD.ARRAY), "GUID_ARRAY", decode_guids
)
"""Array of 128-bit unique identifiers"""

CURRENCY = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.CURRENCY), "CURRENCY", decode_raw
)
"""Windows CURRENCY structure. For now, decoded as raw bytes."""

CURRENCY_ARRAY = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.CURRENCY | DEVPROP_TYPEMOD.ARRAY),
    "CURRENCY_ARRAY",
    decode_raw,
)
"""Array of Windows CURRENCY structures. For now, decoded as raw bytes."""

DATE = PnpPropertyType.register_new(int(DEVPROP_TYPE.DATE), "DATE", decode_raw)
"""64-bit floating-point number (double) that specifies the number of days since December 31, 1899. For now, decoded as raw bytes."""

DATE_ARRAY = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.DATE | DEVPROP_TYPEMOD.ARRAY), "DATE_ARRAY", decode_raw
)
"""Array of doubles that specify the number of days since December 31, 1899. For now, decoded as raw bytes."""

FILETIME = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.FILETIME), "FILETIME", lambda x: one(decode_filetimes(x))
)
"""Windows FILETIME structure."""

FILETIME_ARRAY = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.FILETIME | DEVPROP_TYPEMOD.ARRAY),
    "FILETIME_ARRAY",
    decode_filetimes,
)
"""Array of Windows file time structures."""

BOOLEAN = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.BOOLEAN), "BOOLEAN", lambda x: one(decode_booleans(x))
)
"""8-bit boolean"""

BOOLEAN_ARRAY = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.BOOLEAN | DEVPROP_TYPEMOD.ARRAY), "BOOLEAN_ARRAY", decode_booleans
)
"""Array of 8-bit booleans"""

STRING = PnpPropertyType.register_new(int(DEVPROP_TYPE.STRING), "STRING", decode_string)
"""Null-terminated UTF-16-LE string."""

STRING_LIST = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.STRING | DEVPROP_TYPEMOD.LIST), "STRING_LIST", decode_strings
)
"""Multi-string: sequence of null-terminated UTF-16-LE string, followed by an empty string."""

SECURITY_DESCRIPTOR = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.SECURITY_DESCRIPTOR), "SECURITY_DESCRIPTOR", decode_raw
)
"""Self-relative binary security descriptor. For now, decoded as raw bytes."""

SECURITY_DESCRIPTOR_STRING = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.SECURITY_DESCRIPTOR_STRING),
    "SECURITY_DESCRIPTOR_STRING",
    decode_string,
)
"""Null-terminated UTF-16-LE string that contains a security descriptor in the Security Descriptor Definition Language (SDDL) format."""

SECURITY_DESCRIPTOR_STRING_LIST = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.SECURITY_DESCRIPTOR_STRING | DEVPROP_TYPEMOD.LIST),
    "SECURITY_DESCRIPTOR_STRING_LIST",
    decode_strings,
)
"""Multi-string of strings that contain a security descriptor in the Security Descriptor Definition Language (SDDL) format."""

DEVPROPKEY = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.DEVPROPKEY), "DEVPROPKEY", lambda x: one(decode_property_keys(x))
)
"""Device property key."""

DEVPROPKEY_ARRAY = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.DEVPROPKEY | DEVPROP_TYPEMOD.ARRAY),
    "DEVPROPKEY_ARRAY",
    decode_property_keys,
)
"""Array of device property keys."""

DEVPROPTYPE = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.DEVPROPTYPE),
    "DEVPROPTYPE",
    lambda x: one(decode_property_types(x)),
)
"""Device property type."""

DEVPROPTYPE_ARRAY = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.DEVPROPTYPE | DEVPROP_TYPEMOD.ARRAY),
    "DEVPROPTYPE_ARRAY",
    decode_property_types,
)
"""Array of device property types."""

ERROR = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.ERROR), "ERROR", lambda x: one(decode_win32_errors(x))
)
"""32-bit Win32 system error code."""

ERROR_ARRAY = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.ERROR | DEVPROP_TYPEMOD.ARRAY), "ERROR_ARRAY", decode_win32_errors
)
"""Array of 32-bit Win32 system error codes."""

NTSTATUS = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.NTSTATUS), "NTSTATUS", lambda x: one(decode_nt_statuses(x))
)
"""32-bit NTSTATUS code."""

NTSTATUS_ARRAY = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.NTSTATUS | DEVPROP_TYPEMOD.ARRAY),
    "NTSTATUS_ARRAY",
    decode_nt_statuses,
)
"""Array of 32-bit NTSTATUS codes."""

STRING_INDIRECT = PnpPropertyType.register_new(
    int(DEVPROP_TYPE.STRING_INDIRECT), "STRING_INDIRECT", decode_string
)
"""Null-terminated UTF-16-LE string that contains an indirect string reference."""



from collections.abc import Mapping
from contextlib import AbstractContextManager, nullcontext
from dataclasses import dataclass
from typing import Any, Iterator, Optional, cast
from uuid import UUID

from pytest import raises
from pytest_cases import fixture, parametrize, parametrize_with_cases

from winpnp.info.device import DeviceInfo
from winpnp.info.setup_class import SetupClassInfo
from winpnp.properties import keys, kinds
from winpnp.properties.pnp_property import PnpProperty, PnpPropertyKey, PnpPropertyType

ROOT_INSTANCE_ID = "HTREE\\ROOT\\0"
UNKNOWN_CLASS_GUID = UUID("4d36e97e-e325-11ce-bfc1-08002be10318")
UNKNOWN_CLASS_NAME = "Unknown"

GetitemTestParams = tuple[
    Mapping[PnpPropertyKey[Any], PnpProperty[Any]],
    Any,
    Optional[PnpProperty[Any]],
    AbstractContextManager[Any],
]


@dataclass()
class PnpMappingTestData:
    mapping: Mapping[PnpPropertyKey[Any], PnpProperty[Any]]
    valid_key: PnpPropertyKey[Any]
    expected_value: PnpProperty[Any]


@fixture(scope="function")  # type: ignore
def device() -> Iterator[PnpMappingTestData]:
    # Using the root of the device tree, since it should always exist
    with DeviceInfo.of_instance_id(ROOT_INSTANCE_ID) as d:
        yield PnpMappingTestData(
            d, keys.device.INSTANCE_ID, PnpProperty(ROOT_INSTANCE_ID, kinds.STRING)
        )


@fixture(scope="function")  # type: ignore
def setup_class() -> PnpMappingTestData:
    return PnpMappingTestData(
        SetupClassInfo(UNKNOWN_CLASS_GUID),
        keys.device_class.CLASS_NAME,
        PnpProperty(UNKNOWN_CLASS_NAME, kinds.STRING),
    )


@parametrize("data", (device, setup_class))
def case_success_with_valid_key(
    data: PnpMappingTestData,
) -> GetitemTestParams:
    return data.mapping, data.valid_key, data.expected_value, nullcontext()


@parametrize("data", (device, setup_class))
def case_success_when_allowed_types_not_specified(
    data: PnpMappingTestData,
) -> GetitemTestParams:
    return (
        data.mapping,
        PnpPropertyKey(
            data.valid_key.category, data.valid_key.property_id, allowed_types=None
        ),
        data.expected_value,
        nullcontext(),
    )


@parametrize("data", (device, setup_class))
def case_success_with_multiple_allowed_types(
    data: PnpMappingTestData,
) -> GetitemTestParams:
    return (
        data.mapping,
        PnpPropertyKey(
            data.valid_key.category,
            data.valid_key.property_id,
            allowed_types=(
                cast(PnpPropertyType[Any], kinds.UINT32),
                cast(PnpPropertyType[Any], kinds.INT16_ARRAY),
                data.expected_value.kind,
            ),
        ),
        data.expected_value,
        nullcontext(),
    )


@parametrize("data", (device, setup_class))
def case_raise_value_error_if_return_type_does_not_match_allowed_types(
    data: PnpMappingTestData,
) -> GetitemTestParams:
    return (
        data.mapping,
        PnpPropertyKey(
            data.valid_key.category,
            data.valid_key.property_id,
            allowed_types=(kinds.NULL,),
        ),
        None,
        raises(ValueError),
    )


@parametrize("data", (device, setup_class))
def case_raise_key_error_if_key_is_missing(
    data: PnpMappingTestData,
) -> GetitemTestParams:
    return data.mapping, PnpPropertyKey(UUID(int=0), 0), None, raises(KeyError)


@parametrize("data", (device, setup_class))
def case_raise_key_error_if_key_has_incorrect_type(
    data: PnpMappingTestData,
) -> GetitemTestParams:
    return data.mapping, object(), None, raises(KeyError)


@parametrize_with_cases(
    ("mapping", "key", "expected_result", "expected_exception_context"), cases="."
)
def test_getitem(
    mapping: Mapping[PnpPropertyKey[Any], PnpProperty[Any]],
    key: Any,
    expected_result: Optional[PnpProperty[Any]],
    expected_exception_context: AbstractContextManager[Any],
) -> None:
    actual_result = None

    with expected_exception_context:
        actual_result = mapping[key]

    assert actual_result == expected_result



from uuid import UUID

from winpnp.properties import kinds
from winpnp.properties.pnp_property import PnpPropertyKey

DEVICE_ADDRESS = PnpPropertyKey.register_new(
    UUID("{2bd67d8b-8beb-48d5-87e06cda3428040a}"),
    1,
    "Bluetooth_DeviceAddress",
    (kinds.STRING,),
)

SERVICE_GUID = PnpPropertyKey.register_new(
    UUID("{2bd67d8b-8beb-48d5-87e06cda3428040a}"),
    2,
    "Bluetooth_ServiceGUID",
    (kinds.GUID,),
)

DEVICE_FLAGS = PnpPropertyKey.register_new(
    UUID("{2bd67d8b-8beb-48d5-87e06cda3428040a}"),
    3,
    "Bluetooth_DeviceFlags",
    (kinds.UINT32,),
)

DEVICE_MANUFACTURER = PnpPropertyKey.register_new(
    UUID("{2bd67d8b-8beb-48d5-87e06cda3428040a}"),
    4,
    "Bluetooth_DeviceManufacturer",
    (kinds.STRING,),
)

DEVICE_MODEL_NUMBER = PnpPropertyKey.register_new(
    UUID("{2bd67d8b-8beb-48d5-87e06cda3428040a}"),
    5,
    "Bluetooth_DeviceModelNumber",
    (kinds.STRING,),
)

DEVICE_VID_SOURCE = PnpPropertyKey.register_new(
    UUID("{2bd67d8b-8beb-48d5-87e06cda3428040a}"),
    6,
    "Bluetooth_DeviceVIDSource",
    (kinds.BYTE,),
)

DEVICE_VID = PnpPropertyKey.register_new(
    UUID("{2bd67d8b-8beb-48d5-87e06cda3428040a}"),
    7,
    "Bluetooth_DeviceVID",
    (kinds.UINT16,),
)

DEVICE_PID = PnpPropertyKey.register_new(
    UUID("{2bd67d8b-8beb-48d5-87e06cda3428040a}"),
    8,
    "Bluetooth_DevicePID",
    (kinds.UINT16,),
)

DEVICE_PRODUCT_VERSION = PnpPropertyKey.register_new(
    UUID("{2bd67d8b-8beb-48d5-87e06cda3428040a}"),
    9,
    "Bluetooth_DeviceProductVersion",
    (kinds.UINT16,),
)

CLASS_OF_DEVICE = PnpPropertyKey.register_new(
    UUID("{2bd67d8b-8beb-48d5-87e06cda3428040a}"),
    10,
    "Bluetooth_ClassOfDevice",
    (kinds.UINT32,),
)

LAST_CONNECTED_TIME = PnpPropertyKey.register_new(
    UUID("{2bd67d8b-8beb-48d5-87e06cda3428040a}"),
    11,
    "Bluetooth_LastConnectedTime",
    (kinds.FILETIME,),
)

LAST_SEEN_TIME = PnpPropertyKey.register_new(
    UUID("{2bd67d8b-8beb-48d5-87e06cda3428040a}"),
    12,
    "Bluetooth_LastSeenTime",
    (kinds.FILETIME,),
)



from uuid import UUID

from winpnp.properties import kinds
from winpnp.properties.pnp_property import PnpPropertyKey

from . import (
    dev_query,
    device,
    device_class,
    device_container,
    device_interface,
    device_interface_class,
    driver_package,
)

NAME = PnpPropertyKey.register_new(
    UUID("{b725f130-47ef-101a-a5f102608c9eebac}"), 10, "NAME", (kinds.STRING,)
)



from uuid import UUID

from winpnp.properties import kinds
from winpnp.properties.pnp_property import PnpPropertyKey

DEVICE_DESC = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"),
    2,
    "Device_DeviceDesc",
    (kinds.STRING,),
)

HARDWARE_IDS = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"),
    3,
    "Device_HardwareIds",
    (kinds.STRING_LIST,),
)


COMPATIBLE_IDS = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"),
    4,
    "Device_CompatibleIds",
    (kinds.STRING_LIST,),
)

SERVICE = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"), 6, "Device_Service", (kinds.STRING,)
)

CLASS = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"), 9, "Device_Class", (kinds.STRING,)
)

CLASS_GUID = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"), 10, "Device_ClassGuid", (kinds.GUID,)
)

DRIVER = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"), 11, "Device_Driver", (kinds.STRING,)
)

CONFIG_FLAGS = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"),
    12,
    "Device_ConfigFlags",
    (kinds.UINT32,),
)

MANUFACTURER = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"),
    13,
    "Device_Manufacturer",
    (kinds.STRING,),
)

FRIENDLY_NAME = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"),
    14,
    "Device_FriendlyName",
    (kinds.STRING,),
)

LOCATION_INFO = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"),
    15,
    "Device_LocationInfo",
    (kinds.STRING,),
)

PDO_NAME = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"), 16, "Device_PDOName", (kinds.STRING,)
)

CAPABILITIES = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"),
    17,
    "Device_Capabilities",
    (kinds.UINT32,),
)

UI_NUMBER = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"),
    18,
    "Device_UINumber",
    (kinds.UINT32,),
)

UPPER_FILTERS = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"),
    19,
    "Device_UpperFilters",
    (kinds.STRING_LIST,),
)

LOWER_FILTERS = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"),
    20,
    "Device_LowerFilters",
    (kinds.STRING_LIST,),
)

BUS_TYPE_GUID = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"),
    21,
    "Device_BusTypeGuid",
    (kinds.GUID,),
)

LEGACY_BUS_TYPE = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"),
    22,
    "Device_LegacyBusType",
    (kinds.UINT32,),
)

BUS_NUMBER = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"),
    23,
    "Device_BusNumber",
    (kinds.UINT32,),
)

ENUMERATOR_NAME = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"),
    24,
    "Device_EnumeratorName",
    (kinds.STRING,),
)

SECURITY = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"),
    25,
    "Device_Security",
    (kinds.SECURITY_DESCRIPTOR,),
)

SECURITY_SDS = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"),
    26,
    "Device_SecuritySDS",
    (kinds.SECURITY_DESCRIPTOR_STRING,),
)

DEV_TYPE = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"), 27, "Device_DevType", (kinds.UINT32,)
)

EXCLUSIVE = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"),
    28,
    "Device_Exclusive",
    (kinds.BOOLEAN,),
)

CHARACTERISTICS = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"),
    29,
    "Device_Characteristics",
    (kinds.UINT32,),
)

ADDRESS = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"), 30, "Device_Address", (kinds.UINT32,)
)

UI_NUMBER_DESC_FORMAT = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"),
    31,
    "Device_UINumberDescFormat",
    (kinds.STRING,),
)

POWER_DATA = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"),
    32,
    "Device_PowerData",
    (kinds.BYTE_ARRAY,),
)

REMOVAL_POLICY = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"),
    33,
    "Device_RemovalPolicy",
    (kinds.UINT32,),
)

REMOVAL_POLICY_DEFAULT = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"),
    34,
    "Device_RemovalPolicyDefault",
    (kinds.UINT32,),
)

REMOVAL_POLICY_OVERRIDE = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"),
    35,
    "Device_RemovalPolicyOverride",
    (kinds.UINT32,),
)

INSTALL_STATE = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"),
    36,
    "Device_InstallState",
    (kinds.UINT32,),
)

LOCATION_PATHS = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"),
    37,
    "Device_LocationPaths",
    (kinds.STRING_LIST,),
)

BASE_CONTAINER_ID = PnpPropertyKey.register_new(
    UUID("{a45c254e-df1c-4efd-802067d146a850e0}"),
    38,
    "Device_BaseContainerId",
    (kinds.GUID,),
)

INSTANCE_ID = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    256,
    "Device_InstanceId",
    (kinds.STRING,),
)

DEV_NODE_STATUS = PnpPropertyKey.register_new(
    UUID("{4340a6c5-93fa-4706-972c7b648008a5a7}"),
    2,
    "Device_DevNodeStatus",
    (kinds.UINT32,),
)

PROBLEM_CODE = PnpPropertyKey.register_new(
    UUID("{4340a6c5-93fa-4706-972c7b648008a5a7}"),
    3,
    "Device_ProblemCode",
    (kinds.UINT32,),
)

EJECTION_RELATIONS = PnpPropertyKey.register_new(
    UUID("{4340a6c5-93fa-4706-972c7b648008a5a7}"),
    4,
    "Device_EjectionRelations",
    (kinds.STRING_LIST,),
)

REMOVAL_RELATIONS = PnpPropertyKey.register_new(
    UUID("{4340a6c5-93fa-4706-972c7b648008a5a7}"),
    5,
    "Device_RemovalRelations",
    (kinds.STRING_LIST,),
)

POWER_RELATIONS = PnpPropertyKey.register_new(
    UUID("{4340a6c5-93fa-4706-972c7b648008a5a7}"),
    6,
    "Device_PowerRelations",
    (kinds.STRING_LIST,),
)

BUS_RELATIONS = PnpPropertyKey.register_new(
    UUID("{4340a6c5-93fa-4706-972c7b648008a5a7}"),
    7,
    "Device_BusRelations",
    (kinds.STRING_LIST,),
)

PARENT = PnpPropertyKey.register_new(
    UUID("{4340a6c5-93fa-4706-972c7b648008a5a7}"), 8, "Device_Parent", (kinds.STRING,)
)

CHILDREN = PnpPropertyKey.register_new(
    UUID("{4340a6c5-93fa-4706-972c7b648008a5a7}"),
    9,
    "Device_Children",
    (kinds.STRING_LIST,),
)

SIBLINGS = PnpPropertyKey.register_new(
    UUID("{4340a6c5-93fa-4706-972c7b648008a5a7}"),
    10,
    "Device_Siblings",
    (kinds.STRING_LIST,),
)

TRANSPORT_RELATIONS = PnpPropertyKey.register_new(
    UUID("{4340a6c5-93fa-4706-972c7b648008a5a7}"),
    11,
    "Device_TransportRelations",
    (kinds.STRING_LIST,),
)

PROBLEM_STATUS = PnpPropertyKey.register_new(
    UUID("{4340a6c5-93fa-4706-972c7b648008a5a7}"),
    12,
    "Device_ProblemStatus",
    (kinds.NTSTATUS,),
)

REPORTED = PnpPropertyKey.register_new(
    UUID("{80497100-8c73-48b9-aad9ce387e19c56e}"),
    2,
    "Device_Reported",
    (kinds.BOOLEAN,),
)

LEGACY = PnpPropertyKey.register_new(
    UUID("{80497100-8c73-48b9-aad9ce387e19c56e}"), 3, "Device_Legacy", (kinds.BOOLEAN,)
)

CONTAINER_ID = PnpPropertyKey.register_new(
    UUID("{8c7ed206-3f8a-4827-b3abae9e1faefc6c}"),
    2,
    "Device_ContainerId",
    (kinds.GUID,),
)

IN_LOCAL_MACHINE_CONTAINER = PnpPropertyKey.register_new(
    UUID("{8c7ed206-3f8a-4827-b3abae9e1faefc6c}"),
    4,
    "Device_InLocalMachineContainer",
    (kinds.BOOLEAN,),
)

MODEL = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"), 39, "Device_Model", (kinds.STRING,)
)

MODEL_ID = PnpPropertyKey.register_new(
    UUID("{80d81ea6-7473-4b0c-8216efc11a2c4c8b}"), 2, "Device_ModelId", (kinds.GUID,)
)

FRIENDLY_NAME_ATTRIBUTES = PnpPropertyKey.register_new(
    UUID("{80d81ea6-7473-4b0c-8216efc11a2c4c8b}"),
    3,
    "Device_FriendlyNameAttributes",
    (kinds.UINT32,),
)

MANUFACTURER_ATTRIBUTES = PnpPropertyKey.register_new(
    UUID("{80d81ea6-7473-4b0c-8216efc11a2c4c8b}"),
    4,
    "Device_ManufacturerAttributes",
    (kinds.UINT32,),
)

PRESENCE_NOT_FOR_DEVICE = PnpPropertyKey.register_new(
    UUID("{80d81ea6-7473-4b0c-8216efc11a2c4c8b}"),
    5,
    "Device_PresenceNotForDevice",
    (kinds.BOOLEAN,),
)

SIGNAL_STRENGTH = PnpPropertyKey.register_new(
    UUID("{80d81ea6-7473-4b0c-8216efc11a2c4c8b}"),
    6,
    "Device_SignalStrength",
    (kinds.INT32,),
)

IS_ASSOCIATEABLE_BY_USER_ACTION = PnpPropertyKey.register_new(
    UUID("{80d81ea6-7473-4b0c-8216efc11a2c4c8b}"),
    7,
    "Device_IsAssociateableByUserAction",
    (kinds.BOOLEAN,),
)

SHOW_IN_UNINSTALL_UI = PnpPropertyKey.register_new(
    UUID("{80d81ea6-7473-4b0c-8216efc11a2c4c8b}"),
    8,
    "Device_ShowInUninstallUI",
    (kinds.BOOLEAN,),
)

COMPANION_APPS = PnpPropertyKey.register_new(
    UUID("{6a742654-d0b2-4420-a523e068352ac1df}"),
    2,
    "Device_CompanionApps",
    (kinds.STRING_LIST,),
)

PRIMARY_COMPANION_APP = PnpPropertyKey.register_new(
    UUID("{6a742654-d0b2-4420-a523e068352ac1df}"),
    3,
    "Device_PrimaryCompanionApp",
    (kinds.STRING,),
)

NUMA_PROXIMITY_DOMAIN = PnpPropertyKey.register_new(
    UUID("{540b947e-8b40-45bc-a8a26a0b894cbda2}"),
    1,
    "Device_Numa_Proximity_Domain",
    (kinds.UINT32,),
)

DHP_REBALANCE_POLICY = PnpPropertyKey.register_new(
    UUID("{540b947e-8b40-45bc-a8a26a0b894cbda2}"),
    2,
    "Device_DHP_Rebalance_Policy",
    (kinds.UINT32,),
)

NUMA_NODE = PnpPropertyKey.register_new(
    UUID("{540b947e-8b40-45bc-a8a26a0b894cbda2}"),
    3,
    "Device_Numa_Node",
    (kinds.UINT32,),
)

BUS_REPORTED_DEVICE_DESC = PnpPropertyKey.register_new(
    UUID("{540b947e-8b40-45bc-a8a26a0b894cbda2}"),
    4,
    "Device_BusReportedDeviceDesc",
    (kinds.STRING,),
)

IS_PRESENT = PnpPropertyKey.register_new(
    UUID("{540b947e-8b40-45bc-a8a26a0b894cbda2}"),
    5,
    "Device_IsPresent",
    (kinds.BOOLEAN,),
)

HAS_PROBLEM = PnpPropertyKey.register_new(
    UUID("{540b947e-8b40-45bc-a8a26a0b894cbda2}"),
    6,
    "Device_HasProblem",
    (kinds.BOOLEAN,),
)

CONFIGURATION_ID = PnpPropertyKey.register_new(
    UUID("{540b947e-8b40-45bc-a8a26a0b894cbda2}"),
    7,
    "Device_ConfigurationId",
    (kinds.STRING,),
)

REPORTED_DEVICE_IDS_HASH = PnpPropertyKey.register_new(
    UUID("{540b947e-8b40-45bc-a8a26a0b894cbda2}"),
    8,
    "Device_ReportedDeviceIdsHash",
    (kinds.UINT32,),
)

PHYSICAL_DEVICE_LOCATION = PnpPropertyKey.register_new(
    UUID("{540b947e-8b40-45bc-a8a26a0b894cbda2}"),
    9,
    "Device_PhysicalDeviceLocation",
    (kinds.BYTE_ARRAY,),
)

BIOS_DEVICE_NAME = PnpPropertyKey.register_new(
    UUID("{540b947e-8b40-45bc-a8a26a0b894cbda2}"),
    10,
    "Device_BiosDeviceName",
    (kinds.STRING,),
)

DRIVER_PROBLEM_DESC = PnpPropertyKey.register_new(
    UUID("{540b947e-8b40-45bc-a8a26a0b894cbda2}"),
    11,
    "Device_DriverProblemDesc",
    (kinds.STRING,),
)

DEBUGGER_SAFE = PnpPropertyKey.register_new(
    UUID("{540b947e-8b40-45bc-a8a26a0b894cbda2}"),
    12,
    "Device_DebuggerSafe",
    (kinds.UINT32,),
)

POST_INSTALL_IN_PROGRESS = PnpPropertyKey.register_new(
    UUID("{540b947e-8b40-45bc-a8a26a0b894cbda2}"),
    13,
    "Device_PostInstallInProgress",
    (kinds.BOOLEAN,),
)

STACK = PnpPropertyKey.register_new(
    UUID("{540b947e-8b40-45bc-a8a26a0b894cbda2}"),
    14,
    "Device_Stack",
    (kinds.STRING_LIST,),
)

EXTENDED_CONFIGURATION_IDS = PnpPropertyKey.register_new(
    UUID("{540b947e-8b40-45bc-a8a26a0b894cbda2}"),
    15,
    "Device_ExtendedConfigurationIds",
    (kinds.STRING_LIST,),
)

IS_REBOOT_REQUIRED = PnpPropertyKey.register_new(
    UUID("{540b947e-8b40-45bc-a8a26a0b894cbda2}"),
    16,
    "Device_IsRebootRequired",
    (kinds.BOOLEAN,),
)

FIRMWARE_DATE = PnpPropertyKey.register_new(
    UUID("{540b947e-8b40-45bc-a8a26a0b894cbda2}"),
    17,
    "Device_FirmwareDate",
    (kinds.FILETIME,),
)

FIRMWARE_VERSION = PnpPropertyKey.register_new(
    UUID("{540b947e-8b40-45bc-a8a26a0b894cbda2}"),
    18,
    "Device_FirmwareVersion",
    (kinds.STRING,),
)

FIRMWARE_REVISION = PnpPropertyKey.register_new(
    UUID("{540b947e-8b40-45bc-a8a26a0b894cbda2}"),
    19,
    "Device_FirmwareRevision",
    (kinds.STRING,),
)

DEPENDENCY_PROVIDERS = PnpPropertyKey.register_new(
    UUID("{540b947e-8b40-45bc-a8a26a0b894cbda2}"),
    20,
    "Device_DependencyProviders",
    (kinds.STRING_LIST,),
)

DEPENDENCY_DEPENDENTS = PnpPropertyKey.register_new(
    UUID("{540b947e-8b40-45bc-a8a26a0b894cbda2}"),
    21,
    "Device_DependencyDependents",
    (kinds.STRING_LIST,),
)

SOFT_RESTART_SUPPORTED = PnpPropertyKey.register_new(
    UUID("{540b947e-8b40-45bc-a8a26a0b894cbda2}"),
    22,
    "Device_SoftRestartSupported",
    (kinds.BOOLEAN,),
)

EXTENDED_ADDRESS = PnpPropertyKey.register_new(
    UUID("{540b947e-8b40-45bc-a8a26a0b894cbda2}"),
    23,
    "Device_ExtendedAddress",
    (kinds.UINT64,),
)

ASSIGNED_TO_GUEST = PnpPropertyKey.register_new(
    UUID("{540b947e-8b40-45bc-a8a26a0b894cbda2}"),
    24,
    "Device_AssignedToGuest",
    (kinds.BOOLEAN,),
)

CREATOR_PROCESS_ID = PnpPropertyKey.register_new(
    UUID("{540b947e-8b40-45bc-a8a26a0b894cbda2}"),
    25,
    "Device_CreatorProcessId",
    (kinds.UINT32,),
)

FIRMWARE_VENDOR = PnpPropertyKey.register_new(
    UUID("{540b947e-8b40-45bc-a8a26a0b894cbda2}"),
    26,
    "Device_FirmwareVendor",
    (kinds.STRING,),
)

SESSION_ID = PnpPropertyKey.register_new(
    UUID("{83da6326-97a6-4088-9453a1923f573b29}"),
    6,
    "Device_SessionId",
    (kinds.UINT32,),
)

INSTALL_DATE = PnpPropertyKey.register_new(
    UUID("{83da6326-97a6-4088-9453a1923f573b29}"),
    100,
    "Device_InstallDate",
    (kinds.FILETIME,),
)

FIRST_INSTALL_DATE = PnpPropertyKey.register_new(
    UUID("{83da6326-97a6-4088-9453a1923f573b29}"),
    101,
    "Device_FirstInstallDate",
    (kinds.FILETIME,),
)

LAST_ARRIVAL_DATE = PnpPropertyKey.register_new(
    UUID("{83da6326-97a6-4088-9453a1923f573b29}"),
    102,
    "Device_LastArrivalDate",
    (kinds.FILETIME,),
)

LAST_REMOVAL_DATE = PnpPropertyKey.register_new(
    UUID("{83da6326-97a6-4088-9453a1923f573b29}"),
    103,
    "Device_LastRemovalDate",
    (kinds.FILETIME,),
)

DRIVER_DATE = PnpPropertyKey.register_new(
    UUID("{a8b865dd-2e3d-4094-ad97e593a70c75d6}"),
    2,
    "Device_DriverDate",
    (kinds.FILETIME,),
)

DRIVER_VERSION = PnpPropertyKey.register_new(
    UUID("{a8b865dd-2e3d-4094-ad97e593a70c75d6}"),
    3,
    "Device_DriverVersion",
    (kinds.STRING,),
)

DRIVER_DESC = PnpPropertyKey.register_new(
    UUID("{a8b865dd-2e3d-4094-ad97e593a70c75d6}"),
    4,
    "Device_DriverDesc",
    (kinds.STRING,),
)

DRIVER_INF_PATH = PnpPropertyKey.register_new(
    UUID("{a8b865dd-2e3d-4094-ad97e593a70c75d6}"),
    5,
    "Device_DriverInfPath",
    (kinds.STRING,),
)

DRIVER_INF_SECTION = PnpPropertyKey.register_new(
    UUID("{a8b865dd-2e3d-4094-ad97e593a70c75d6}"),
    6,
    "Device_DriverInfSection",
    (kinds.STRING,),
)

DRIVER_INF_SECTION_EXT = PnpPropertyKey.register_new(
    UUID("{a8b865dd-2e3d-4094-ad97e593a70c75d6}"),
    7,
    "Device_DriverInfSectionExt",
    (kinds.STRING,),
)

MATCHING_DEVICE_ID = PnpPropertyKey.register_new(
    UUID("{a8b865dd-2e3d-4094-ad97e593a70c75d6}"),
    8,
    "Device_MatchingDeviceId",
    (kinds.STRING,),
)

DRIVER_PROVIDER = PnpPropertyKey.register_new(
    UUID("{a8b865dd-2e3d-4094-ad97e593a70c75d6}"),
    9,
    "Device_DriverProvider",
    (kinds.STRING,),
)

DRIVER_PROP_PAGE_PROVIDER = PnpPropertyKey.register_new(
    UUID("{a8b865dd-2e3d-4094-ad97e593a70c75d6}"),
    10,
    "Device_DriverPropPageProvider",
    (kinds.STRING,),
)

DRIVER_CO_INSTALLERS = PnpPropertyKey.register_new(
    UUID("{a8b865dd-2e3d-4094-ad97e593a70c75d6}"),
    11,
    "Device_DriverCoInstallers",
    (kinds.STRING_LIST,),
)

RESOURCE_PICKER_TAGS = PnpPropertyKey.register_new(
    UUID("{a8b865dd-2e3d-4094-ad97e593a70c75d6}"),
    12,
    "Device_ResourcePickerTags",
    (kinds.STRING,),
)

RESOURCE_PICKER_EXCEPTIONS = PnpPropertyKey.register_new(
    UUID("{a8b865dd-2e3d-4094-ad97e593a70c75d6}"),
    13,
    "Device_ResourcePickerExceptions",
    (kinds.STRING,),
)

DRIVER_RANK = PnpPropertyKey.register_new(
    UUID("{a8b865dd-2e3d-4094-ad97e593a70c75d6}"),
    14,
    "Device_DriverRank",
    (kinds.UINT32,),
)

DRIVER_LOGO_LEVEL = PnpPropertyKey.register_new(
    UUID("{a8b865dd-2e3d-4094-ad97e593a70c75d6}"),
    15,
    "Device_DriverLogoLevel",
    (kinds.UINT32,),
)

NO_CONNECT_SOUND = PnpPropertyKey.register_new(
    UUID("{a8b865dd-2e3d-4094-ad97e593a70c75d6}"),
    17,
    "Device_NoConnectSound",
    (kinds.BOOLEAN,),
)

GENERIC_DRIVER_INSTALLED = PnpPropertyKey.register_new(
    UUID("{a8b865dd-2e3d-4094-ad97e593a70c75d6}"),
    18,
    "Device_GenericDriverInstalled",
    (kinds.BOOLEAN,),
)

ADDITIONAL_SOFTWARE_REQUESTED = PnpPropertyKey.register_new(
    UUID("{a8b865dd-2e3d-4094-ad97e593a70c75d6}"),
    19,
    "Device_AdditionalSoftwareRequested",
    (kinds.BOOLEAN,),
)

SAFE_REMOVAL_REQUIRED = PnpPropertyKey.register_new(
    UUID("{afd97640-86a3-4210-b67c289c41aabe55}"),
    2,
    "Device_SafeRemovalRequired",
    (kinds.BOOLEAN,),
)

SAFE_REMOVAL_REQUIRED_OVERRIDE = PnpPropertyKey.register_new(
    UUID("{afd97640-86a3-4210-b67c289c41aabe55}"),
    3,
    "Device_SafeRemovalRequiredOverride",
    (kinds.BOOLEAN,),
)



from collections.abc import Mapping
from ctypes import (
    POINTER,
    GetLastError,
    WinError,
    byref,
    c_ubyte,
    c_uint32,
    cast,
    create_string_buffer,
)
from typing import Any, Iterator, Protocol, TypeVar

from winerror import ERROR_INSUFFICIENT_BUFFER, ERROR_NOT_FOUND

from winpnp._setupapi import DEVPROPKEY, GUID
from winpnp.properties.pnp_property import PnpProperty, PnpPropertyKey, PnpPropertyType

T = TypeVar("T")


class QueryValueFunc(Protocol):
    def __call__(
        self,
        PropertyKey,
        PropertyType,
        PropertyBuffer,
        PropertyBufferSize,
        RequiredSize,
    ) -> bool: ...


class QueryKeysFunc(Protocol):
    def __call__(
        self, PropertyKeyArray, PropertyKeyCount, RequiredPropertyKeyCount
    ) -> bool: ...


class PnpPropertyMapping(Mapping[PnpPropertyKey[T], PnpProperty[T]]):
    def __init__(
        self,
        description: str,
        query_value_func: QueryValueFunc,
        query_keys_func: QueryKeysFunc,
    ) -> None:
        super().__init__()

        self.description = description
        self.query_value = query_value_func
        self.query_keys = query_keys_func

    def __getitem__(self, key: PnpPropertyKey[T]) -> PnpProperty[T]:
        if not isinstance(key, PnpPropertyKey):
            raise KeyError(key)

        win_prop_key = DEVPROPKEY(GUID.from_uuid(key.category), key.property_id)
        type_id = c_uint32()
        required_size = c_uint32()

        if self.query_value(
            byref(win_prop_key),
            byref(type_id),
            None,
            0,
            byref(required_size),
        ):
            return self.__build_property(key, PnpPropertyType(type_id.value), b"")

        error = GetLastError()
        if error == ERROR_NOT_FOUND:
            raise KeyError(key)
        if error != ERROR_INSUFFICIENT_BUFFER:
            raise WinError(error)

        value_buffer = create_string_buffer(required_size.value)
        if not self.query_value(
            byref(win_prop_key),
            byref(type_id),
            cast(value_buffer, POINTER(c_ubyte)),
            required_size,
            None,
        ):
            raise WinError()

        return self.__build_property(
            key, PnpPropertyType(type_id.value), value_buffer.raw
        )

    def __len__(self) -> int:
        required_count = c_uint32()
        if self.query_keys(None, 0, byref(required_count)):
            return 0

        error = GetLastError()
        if error != ERROR_INSUFFICIENT_BUFFER:
            raise WinError(error)

        return required_count.value

    def __iter__(self) -> Iterator[PnpPropertyKey[T]]:
        keys = (DEVPROPKEY * len(self))()
        if not self.query_keys(keys, len(keys), None):
            raise WinError()

        for key in keys:
            yield PnpPropertyKey(key.fmtid.to_uuid(), key.pid)

    def __build_property(
        self,
        key: PnpPropertyKey[T],
        actual_type: PnpPropertyType[Any],
        data: bytes,
    ) -> PnpProperty[T]:
        if (
            key.allowed_types is not None
            and actual_type.type_id not in key.allowed_types
        ):
            raise ValueError(
                f"Property {key} of {self.description} has type {actual_type}, which is not expected."
            )

        if key.allowed_types is not None:
            # Prefer key from allowed_types over original actual_type in case the one from allowed_types has a custom decoder
            actual_type = key.allowed_types[actual_type.type_id]

        return PnpProperty(actual_type.decode(data), actual_type)



from typing import Union, cast
from uuid import UUID

from winpnp.properties import kinds
from winpnp.properties.pnp_property import PnpPropertyKey, PnpPropertyType

ADDRESS = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    51,
    "DeviceContainer_Address",
    (
        cast(PnpPropertyType[Union[str, list[str]]], kinds.STRING),
        cast(PnpPropertyType[Union[str, list[str]]], kinds.STRING_LIST),
    ),
)

DISCOVERY_METHOD = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    52,
    "DeviceContainer_DiscoveryMethod",
    (kinds.STRING_LIST,),
)

IS_ENCRYPTED = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    53,
    "DeviceContainer_IsEncrypted",
    (kinds.BOOLEAN,),
)

IS_AUTHENTICATED = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    54,
    "DeviceContainer_IsAuthenticated",
    (kinds.BOOLEAN,),
)

IS_CONNECTED = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    55,
    "DeviceContainer_IsConnected",
    (kinds.BOOLEAN,),
)

IS_PAIRED = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    56,
    "DeviceContainer_IsPaired",
    (kinds.BOOLEAN,),
)

ICON = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    57,
    "DeviceContainer_Icon",
    (kinds.STRING,),
)

VERSION = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    65,
    "DeviceContainer_Version",
    (kinds.STRING,),
)

LAST_SEEN = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    66,
    "DeviceContainer_Last_Seen",
    (kinds.FILETIME,),
)

LAST_CONNECTED = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    67,
    "DeviceContainer_Last_Connected",
    (kinds.FILETIME,),
)

IS_SHOW_IN_DISCONNECTED_STATE = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    68,
    "DeviceContainer_IsShowInDisconnectedState",
    (kinds.BOOLEAN,),
)

IS_LOCAL_MACHINE = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    70,
    "DeviceContainer_IsLocalMachine",
    (kinds.BOOLEAN,),
)

METADATA_PATH = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    71,
    "DeviceContainer_MetadataPath",
    (kinds.STRING,),
)

IS_METADATA_SEARCH_IN_PROGRESS = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    72,
    "DeviceContainer_IsMetadataSearchInProgress",
    (kinds.BOOLEAN,),
)

METADATA_CHECKSUM = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    73,
    "DeviceContainer_MetadataChecksum",
    (kinds.BYTE_ARRAY,),
)

IS_NOT_INTERESTING_FOR_DISPLAY = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    74,
    "DeviceContainer_IsNotInterestingForDisplay",
    (kinds.BOOLEAN,),
)

LAUNCH_DEVICE_STAGE_ON_DEVICE_CONNECT = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    76,
    "DeviceContainer_LaunchDeviceStageOnDeviceConnect",
    (kinds.BOOLEAN,),
)

LAUNCH_DEVICE_STAGE_FROM_EXPLORER = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    77,
    "DeviceContainer_LaunchDeviceStageFromExplorer",
    (kinds.BOOLEAN,),
)

BASELINE_EXPERIENCE_ID = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    78,
    "DeviceContainer_BaselineExperienceId",
    (kinds.GUID,),
)

IS_DEVICE_UNIQUELY_IDENTIFIABLE = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    79,
    "DeviceContainer_IsDeviceUniquelyIdentifiable",
    (kinds.BOOLEAN,),
)

ASSOCIATION_ARRAY = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    80,
    "DeviceContainer_AssociationArray",
    (kinds.STRING_LIST,),
)

DEVICE_DESCRIPTION1 = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    81,
    "DeviceContainer_DeviceDescription1",
    (kinds.STRING,),
)

DEVICE_DESCRIPTION2 = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    82,
    "DeviceContainer_DeviceDescription2",
    (kinds.STRING,),
)

HAS_PROBLEM = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    83,
    "DeviceContainer_HasProblem",
    (kinds.BOOLEAN,),
)

IS_SHARED_DEVICE = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    84,
    "DeviceContainer_IsSharedDevice",
    (kinds.BOOLEAN,),
)

IS_NETWORK_DEVICE = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    85,
    "DeviceContainer_IsNetworkDevice",
    (kinds.BOOLEAN,),
)

IS_DEFAULT_DEVICE = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    86,
    "DeviceContainer_IsDefaultDevice",
    (kinds.BOOLEAN,),
)

METADATA_CABINET = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    87,
    "DeviceContainer_MetadataCabinet",
    (kinds.STRING,),
)

REQUIRES_PAIRING_ELEVATION = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    88,
    "DeviceContainer_RequiresPairingElevation",
    (kinds.BOOLEAN,),
)

EXPERIENCE_ID = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    89,
    "DeviceContainer_ExperienceId",
    (kinds.GUID,),
)

CATEGORY = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    90,
    "DeviceContainer_Category",
    (kinds.STRING_LIST,),
)

CATEGORY_DESC_SINGULAR = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    91,
    "DeviceContainer_Category_Desc_Singular",
    (kinds.STRING_LIST,),
)

CATEGORY_DESC_PLURAL = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    92,
    "DeviceContainer_Category_Desc_Plural",
    (kinds.STRING_LIST,),
)

CATEGORY_ICON = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    93,
    "DeviceContainer_Category_Icon",
    (kinds.STRING,),
)

CATEGORY_GROUP_DESC = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    94,
    "DeviceContainer_CategoryGroup_Desc",
    (kinds.STRING_LIST,),
)

CATEGORY_GROUP_ICON = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    95,
    "DeviceContainer_CategoryGroup_Icon",
    (kinds.STRING,),
)

PRIMARY_CATEGORY = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    97,
    "DeviceContainer_PrimaryCategory",
    (kinds.STRING,),
)

UNPAIR_UNINSTALL = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    98,
    "DeviceContainer_UnpairUninstall",
    (kinds.BOOLEAN,),
)

REQUIRES_UNINSTALL_ELEVATION = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    99,
    "DeviceContainer_RequiresUninstallElevation",
    (kinds.BOOLEAN,),
)

DEVICE_FUNCTION_SUB_RANK = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    100,
    "DeviceContainer_DeviceFunctionSubRank",
    (kinds.UINT32,),
)

ALWAYS_SHOW_DEVICE_AS_CONNECTED = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    101,
    "DeviceContainer_AlwaysShowDeviceAsConnected",
    (kinds.BOOLEAN,),
)

CONFIG_FLAGS = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    105,
    "DeviceContainer_ConfigFlags",
    (kinds.UINT32,),
)

PRIVILEGED_PACKAGE_FAMILY_NAMES = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    106,
    "DeviceContainer_PrivilegedPackageFamilyNames",
    (kinds.STRING_LIST,),
)

CUSTOM_PRIVILEGED_PACKAGE_FAMILY_NAMES = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    107,
    "DeviceContainer_CustomPrivilegedPackageFamilyNames",
    (kinds.STRING_LIST,),
)

IS_REBOOT_REQUIRED = PnpPropertyKey.register_new(
    UUID("{78c34fc8-104a-4aca-9ea4524d52996e57}"),
    108,
    "DeviceContainer_IsRebootRequired",
    (kinds.BOOLEAN,),
)

FRIENDLY_NAME = PnpPropertyKey.register_new(
    UUID("{656A3BB3-ECC0-43FD-84774AE0404A96CD}"),
    12288,
    "DeviceContainer_FriendlyName",
    (kinds.STRING,),
)

MANUFACTURER = PnpPropertyKey.register_new(
    UUID("{656A3BB3-ECC0-43FD-84774AE0404A96CD}"),
    8192,
    "DeviceContainer_Manufacturer",
    (kinds.STRING,),
)

MODEL_NAME = PnpPropertyKey.register_new(
    UUID("{656A3BB3-ECC0-43FD-84774AE0404A96CD}"),
    8194,
    "DeviceContainer_ModelName",
    (kinds.STRING,),
)

MODEL_NUMBER = PnpPropertyKey.register_new(
    UUID("{656A3BB3-ECC0-43FD-84774AE0404A96CD}"),
    8195,
    "DeviceContainer_ModelNumber",
    (kinds.STRING,),
)

INSTALL_IN_PROGRESS = PnpPropertyKey.register_new(
    UUID("{83da6326-97a6-4088-9453a1923f573b29}"),
    9,
    "DeviceContainer_InstallInProgress",
    (kinds.BOOLEAN,),
)



from uuid import UUID

from winpnp.properties import kinds
from winpnp.properties.pnp_property import PnpPropertyKey

OBJECT_TYPE = PnpPropertyKey.register_new(
    UUID("{13673f42-a3d6-49f6-b4daae46e0c5237c}"),
    2,
    "DevQuery_ObjectType",
    (kinds.UINT32,),
)



from uuid import UUID

from winpnp.properties import kinds
from winpnp.properties.pnp_property import PnpPropertyKey

UPPER_FILTERS = PnpPropertyKey.register_new(
    UUID("{4321918b-f69e-470d-a5de4d88c75ad24b}"),
    19,
    "DeviceClass_UpperFilters",
    (kinds.STRING_LIST,),
)

LOWER_FILTERS = PnpPropertyKey.register_new(
    UUID("{4321918b-f69e-470d-a5de4d88c75ad24b}"),
    20,
    "DeviceClass_LowerFilters",
    (kinds.STRING_LIST,),
)

SECURITY = PnpPropertyKey.register_new(
    UUID("{4321918b-f69e-470d-a5de4d88c75ad24b}"),
    25,
    "DeviceClass_Security",
    (kinds.SECURITY_DESCRIPTOR,),
)

SECURITY_SDS = PnpPropertyKey.register_new(
    UUID("{4321918b-f69e-470d-a5de4d88c75ad24b}"),
    26,
    "DeviceClass_SecuritySDS",
    (kinds.SECURITY_DESCRIPTOR_STRING,),
)

DEV_TYPE = PnpPropertyKey.register_new(
    UUID("{4321918b-f69e-470d-a5de4d88c75ad24b}"),
    27,
    "DeviceClass_DevType",
    (kinds.UINT32,),
)

EXCLUSIVE = PnpPropertyKey.register_new(
    UUID("{4321918b-f69e-470d-a5de4d88c75ad24b}"),
    28,
    "DeviceClass_Exclusive",
    (kinds.BOOLEAN,),
)

CHARACTERISTICS = PnpPropertyKey.register_new(
    UUID("{4321918b-f69e-470d-a5de4d88c75ad24b}"),
    29,
    "DeviceClass_Characteristics",
    (kinds.UINT32,),
)

NAME = PnpPropertyKey.register_new(
    UUID("{259abffc-50a7-47ce-af0868c9a7d73366}"),
    2,
    "DeviceClass_Name",
    (kinds.STRING,),
)

CLASS_NAME = PnpPropertyKey.register_new(
    UUID("{259abffc-50a7-47ce-af0868c9a7d73366}"),
    3,
    "DeviceClass_ClassName",
    (kinds.STRING,),
)

ICON = PnpPropertyKey.register_new(
    UUID("{259abffc-50a7-47ce-af0868c9a7d73366}"),
    4,
    "DeviceClass_Icon",
    (kinds.STRING,),
)

CLASS_INSTALLER = PnpPropertyKey.register_new(
    UUID("{259abffc-50a7-47ce-af0868c9a7d73366}"),
    5,
    "DeviceClass_ClassInstaller",
    (kinds.STRING,),
)

PROP_PAGE_PROVIDER = PnpPropertyKey.register_new(
    UUID("{259abffc-50a7-47ce-af0868c9a7d73366}"),
    6,
    "DeviceClass_PropPageProvider",
    (kinds.STRING,),
)

NO_INSTALL_CLASS = PnpPropertyKey.register_new(
    UUID("{259abffc-50a7-47ce-af0868c9a7d73366}"),
    7,
    "DeviceClass_NoInstallClass",
    (kinds.BOOLEAN,),
)

NO_DISPLAY_CLASS = PnpPropertyKey.register_new(
    UUID("{259abffc-50a7-47ce-af0868c9a7d73366}"),
    8,
    "DeviceClass_NoDisplayClass",
    (kinds.BOOLEAN,),
)

SILENT_INSTALL = PnpPropertyKey.register_new(
    UUID("{259abffc-50a7-47ce-af0868c9a7d73366}"),
    9,
    "DeviceClass_SilentInstall",
    (kinds.BOOLEAN,),
)

NO_USE_CLASS = PnpPropertyKey.register_new(
    UUID("{259abffc-50a7-47ce-af0868c9a7d73366}"),
    10,
    "DeviceClass_NoUseClass",
    (kinds.BOOLEAN,),
)

DEFAULT_SERVICE = PnpPropertyKey.register_new(
    UUID("{259abffc-50a7-47ce-af0868c9a7d73366}"),
    11,
    "DeviceClass_DefaultService",
    (kinds.STRING,),
)

ICON_PATH = PnpPropertyKey.register_new(
    UUID("{259abffc-50a7-47ce-af0868c9a7d73366}"),
    12,
    "DeviceClass_IconPath",
    (kinds.STRING_LIST,),
)

DHP_REBALANCE_OPT_OUT = PnpPropertyKey.register_new(
    UUID("{d14d3ef3-66cf-4ba2-9d380ddb37ab4701}"),
    2,
    "DeviceClass_DHPRebalanceOptOut",
    (kinds.BOOLEAN,),
)

CLASS_CO_INSTALLERS = PnpPropertyKey.register_new(
    UUID("{713d1703-a2e2-49f5-921456472ef3da5c}"),
    2,
    "DeviceClass_ClassCoInstallers",
    (kinds.STRING_LIST,),
)
